#
# Utility to transcode the Zeitgeist ontology from Python code
# to a C header file defining a collection of macros for the
# members of the ontology
#

import sys
import re
import zeitgeist.datamodel as dm

def caseconv(name):
	"""
	Converts CamelCase to CAMEL_CASE
	"""
	result = ""
	for i in range(len(name) - 1) :
		if name[i].islower() and name[i+1].isupper():
			result += name[i].upper() + "_"
		else:
			result += name[i].upper()
	result += name[-1].upper()
	return result

def symbolname (symbol):
	try:
		ns, name = symbol[symbol.rfind("/")+1:].split("#")
		return "ZEITGEIST_%s_%s" % (ns.upper(), caseconv(name))
	except Exception, e:
		return symbol.uri.rpartition("#")[2].upper()

def output_symbol_func (body):
	header = """
static void
_ensure_symbols_loaded (void)
{
  gchar **parents;
  gchar **children;
  gchar **all_children;
  gchar *uri;
  SymbolInfo *info;

  if (symbols_loaded) return;

%s

  symbols_loaded = TRUE;
}
"""

	# indent
	symbols_func = header[1:] % re.sub(r'(?m)^(.+)$', r'  \1', body)
	return symbols_func

def get_symbols (l):
	s = filter(lambda x: x not in [dm.Interpretation, dm.Manifestation], l)
	s = map(symbolname, s)
	s.append("NULL")

	return s

def alloc_array (name, symbols):
	num_elems = len(symbols)
	alloc_call = '%s = g_new (char*, %d);' % (name, num_elems)
	elems = ['%s[%d] = %s;' % (name,i,symbols[i]) for i in xrange(0, num_elems)]
	elems.insert(0, alloc_call)
	return '\n'.join(elems)

def register_symbol (symbol):
	uri = 'uri = %s;' % symbolname(symbol)
	arrays = ['parents', 'children', 'all_children']
	parents = alloc_array(arrays[0], get_symbols(symbol.get_parents()))
	children = alloc_array(arrays[1], get_symbols(symbol.get_children()))
	all_ch = alloc_array(arrays[2], get_symbols(symbol.get_all_children()))

	si_call = 'info = symbol_info_new (uri, parents, children, all_children);'
	register_call = 'zeitgeist_register_symbol (info);'

	calls = [uri, parents, children, all_ch, si_call, register_call]
	calls.extend(['g_free (%s);' % var for var in arrays])

	return '\n'.join(calls) + '\n'

if __name__ == "__main__":
	interpretations = []
	manifestations = []
	for interp in dm.Interpretation.get_all_children():
		try:
			stmt = register_symbol (interp)
			interpretations.append(stmt)
		except Exception, e:
			print >> sys.stderr, "Failed to convert %s: %s" % (interp, e)

	for manif in dm.Manifestation.get_all_children():
		try:
			stmt = register_symbol (manif)
			manifestations.append(stmt)
		except Exception, e:
			print >> sys.stderr, "Failed to convert %s: %s" % (manif, e)


	all_symbols = []
	all_symbols.extend(interpretations)
	all_symbols.extend(manifestations)

	f = open('zeitgeist-symbols.template', 'r')
	print f.read()
	print output_symbol_func('\n'.join(all_symbols))
